/*
Copyright (C) 1999/2000 Hendrik Lipka.
All rights reserved.

This program is free software; you can redistribute it and/or modify it under the terms 
of the GNU General Public License as published by the Free Software Foundation; 
either version 2 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY 
WARRANTY; without even the implied warranty of MERCHANTABILITY or 
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License 
for more details. 

You should have received a copy of the GNU General Public License along with this 
program; if not, write to the Free Software Foundation, Inc., 675 Mass Ave, 
Cambridge, MA 02139, USA or see http://www.gnu.org

If you encounter problems or have a suggestion, please mail to:
hendrik.lipka@gmx.de
*/

package de.hendriklipka.ldse.common;

import java.io.FilterReader;
import java.io.Reader;
import java.io.IOException;
import java.util.Hashtable;

import de.hendriklipka.ldse.common.Log;

/**
  Filtert einen Stream nach HTML-Tags. Das Ergebnis weist nur noch den reinen Text auf.
  Es werden einige besondere Tags gemerkt:
  Titel, Metatags (Autor, Keywords, Content, Dublin Core)
  
  Das Parsing wird mit einer State Machine erledigt
*/
public class HTMLTagFilterReader extends FilterReader
{
  protected String _buffer=null;
  
  protected String _title="";
  protected String _charSet=null;
  protected String _author="";
  protected String _keywords="";
  protected String _description="";
  protected String _dublinCore="";
  
  protected boolean _headerFinished=false;
  protected boolean _inTitle=false;
  
	public HTMLTagFilterReader(Reader in)
	{
		super(in);
	}
	
  public int read(char cbuf[], int off, int len) throws IOException 
  {
    int max=off+len;
    for (int i=off;i<max;i++)
    {
      int c=read();
      if (-1!=c)
        cbuf[i]=(char)c;
      else
        return -1;
    }
    return len;
  }
  
  public int read(char cbuf[]) throws IOException 
  {
    return read(cbuf, 0, cbuf.length);
  }

  private static final int maxSkipBufferSize = 8192;
  private char skipBuffer[] = null;
  public long skip(long n) throws IOException 
  {
    int nn = (int) Math.min(n, maxSkipBufferSize);
    synchronized (lock) 
    {
      if ((skipBuffer == null) || (skipBuffer.length < nn))
        skipBuffer = new char[nn];
      long r = n;
      while (r > 0) 
      {
        int nc = read(skipBuffer, 0, nn);
        if (nc == -1)
          break;
        r -= nc;
      }
      return n - r;
    }
  }
  
  public int read() throws IOException 
  {
    // Puffer, falls was ersetzt werden musste
    // wird nicht noch einmal geparst
    if (null!=_buffer)
    {
      char c=_buffer.charAt(0);
      _buffer=_buffer.substring(1);
      if (0==_buffer.length())
        _buffer=null;
      return c;
    }
    int c=in.read();
    if (-1==c)
      return -1;
    if('<'==c)
    {
      StringBuffer tagName=null;
      StringBuffer tagContent=null;
      if(!_headerFinished)
      {
        tagName=new StringBuffer();
        tagContent=new StringBuffer();
      }
      boolean tagReady=false;
      // alles bis zum schliessenden Tag ueberlesen
      //ToDo: Anfuehrungszeichen???
      // sind die hier erlaubt??
      while ('>'!=c)
      {
        c=in.read();
        if (-1==c)
          return -1;
        if(!_headerFinished)
        {
          if(!tagReady)
          {
            if(' '!=c)
            {
              if('>'!=c)
                tagName.append((char)c);
            }
            else
            {
              tagReady=true;
            }
          }
          else
          {
            if('>'!=c)
              tagContent.append((char)c);
          }
        }
      }
      c=' '; // Tags trennen Worte, also liefern wir nach dem Ende ein SPACE
      // Wenn nach diesem Tag direkt ein weiteres kommt, wird hier ein Space geliefert
      // und das Tag wird beim naechsten Aufruf ueberlesen
      
      // Behandlung fuer einige Tags
      if (!_headerFinished)
      {
        String name=tagName.toString();
        String content=tagContent.toString();
        if (name.equalsIgnoreCase("/HEAD"))
        {
          _headerFinished=true;
        }
        else if(name.equalsIgnoreCase("/TITLE"))
        {
          _inTitle=false;
          return ' ';
        }
        else if(name.equalsIgnoreCase("TITLE"))
        {
          // Beginn des Dokumententitels
          _inTitle=true;
          StringBuffer title=new StringBuffer();
          // lesen wir solange Zeichen, bis (rekursiv) </TITLE> gefunden wurde
          while(_inTitle)
          {
            c=this.read();
            if ((c!='\n')&&(c!='\r'))
              title.append((char)c);
          }
          _title=title.toString();
          // damit liefern wir den Title einfach mit zurueck
          _buffer=_title;
          return this.read();
        }
        else if(name.equalsIgnoreCase("META"))
        {
          // Metatags, die geparst werden muessen
          parseMetaTag(tagContent.toString());
        }
      }
    }
    if('&'==c)
    {
      int counter=0;
      boolean broken=false;
      c=in.read();
      StringBuffer sc=new StringBuffer();
      // bis zum Semikolon lesen
      // Sicherheitsbegrenzung fuer fehlerhaftes HTML:
      // wenn nach dem & ein SPACE kommt oder mehr als 20 Zeichen bis zum ; oder
      // ein Dateiende, wird abgebrochen
      while (true)
      {
        sc.append((char)c);
        c=in.read();
        if(';'==c)
        {
          break;
        }
        else if((' '==c)||(counter>30)||(-1==c))
        {
          _buffer+='&';
          _buffer+=sc;
          broken=true;
          break;
        }
        counter++;
      }
      if (!broken)
      {
        // nach dem Ersetzungszeichen suchen
        String scs=sc.toString();
        if ('#'==scs.charAt(0))
        {
          // Unicode-Character wurde direkt angegeben
          try
          {
            scs=scs.substring(1);
            int cnum=Integer.parseInt(scs);
            c=cnum;
          }
          catch (Exception ex)
          {
            _buffer+='&';
            _buffer+='#';
            _buffer+=scs;
            _buffer+=';';
          }
        }
        else
        {
          Character C=(Character)_specialCharMap.get(scs);
          if (null!=C)
          {
            // anhaengen passiert weiter unten
            c=C.charValue();
          }
          else
          {
            _buffer+='&';
            _buffer+=scs;
            _buffer+=';';
          }
        }
      }
    }
    // wenn der Puffer inzwischen gefuellt wurde, lesen wir ihn einfach aus
    if (null!=_buffer)
      return this.read();
    return c;
  }
  
  protected void parseMetaTag(String tagContent)
  {
    // in diesem String machen wir indexOf(), weil dieses kein ignoreCase kennt
    String s=tagContent.toLowerCase();
    
    String name=null;
    String content=null;

    int nameStart=-1;
    int nameEnd=-1;
    int contentStart=-1;
    int contentEnd=-1;
    
    // parsing
    int pos=0;
    char c=' ';
    String s1="";
    try
    {
      while (pos<tagContent.length())
      {
        // fuehrende Leerzeichen ueberlesen
        c=s.charAt(pos++);
        if(' '==c)
          continue;
        // identifier einlesen
        // am Anfang duerfen hier keine " auftreten
        while((' '!=c)&&('='!=c)&&('"'!=c))
        {
          s1=s1+=c;
          c=s.charAt(pos);
          pos++;
        }
        // jetzt haben wir den ersten identifier
        while ('"'!=c)
        {
          c=s.charAt(pos);
          pos++;
        }
//        pos++;
        if(s1.equals("name")||s1.equals("http-equiv"))
        {
          nameStart=pos;
        }
        else
        {
          contentStart=pos;
        }
        c=s.charAt(pos);
        while ('"'!=c)
        {
          pos++;
          c=s.charAt(pos);
        }
        if(s1.equals("name")||s1.equals("http-equiv"))
        {
          nameEnd=pos;
        }
        else
        {
          contentEnd=pos;
        }
        pos++;
      }
    }
    catch (Exception ex) {}
    
    if((0<=nameStart)&&(0<=nameEnd)&&(nameStart<=nameEnd)&&(0<=contentStart)&&(0<=contentEnd)&&(contentStart<=contentEnd))
    {
      name=tagContent.substring(nameStart,nameEnd);
      content=tagContent.substring(contentStart,contentEnd);
      if (name.equalsIgnoreCase("author"))
      {
        _author=content;
      }
      else if (name.equalsIgnoreCase("keywords"))
      {
        _keywords+=" "+content;
      }
      else if (name.equalsIgnoreCase("description"))
      {
        _description+=" "+content;
      }
      else if (name.startsWith("DC."))
      {
        _dublinCore+=" "+content;
      }
      else if (name.equalsIgnoreCase("content-type"))
      {
        s=content.toLowerCase();
        pos=s.indexOf("charset");
        if (-1<pos)
        {
          pos=s.indexOf('=',pos+1);
          if (-1<pos)
          {
            int pos2=s.indexOf('"',pos+1);
            if (-1<pos2)
            {
              _charSet=content.substring(pos+1,pos2);
            }
            else
            {
              _charSet=content.substring(pos+1);
            }
          }
        }
      }
    }
  }
  
  public void close() throws IOException
  {
    in.close();
  }
  
  public boolean ready() throws IOException 
  {
    return in.ready();
  }
  
  public String getTitle()
  {
    return _title;
  }
  
  public String getCharset()
  {
    return _charSet;
  }
  
  public String getAuthor()
  {
    return _author;
  }
  
  public String getKeywords()
  {
    return _keywords;
  }
  
  public String getDescription()
  {
    return _description;
  }
  
  public String getDublinCore()
  {
    return _dublinCore;
  }
  
  
  protected static final String[][] _specialChars= { // HTML 3.2
                                        {"nbsp"," "},
                                        {"gt",">"},
                                        {"lt","<"},
                                        {"amp","&"},
                                        {"quot","\""},
                                        {"iexcl","\u00a1"},
                                        {"cent","\u00a2"},
                                        {"pound","\u00a3"},
                                        {"curren","\u00a4"},
                                        {"yen","\u00a5"},
                                        {"brvbar","\u00a6"},
                                        {"sect","\u00a7"},
                                        {"uml","\u00a8"},
                                        {"copy","\u00a9"},
                                        {"ordf","\u00aa"},
                                        {"laquo","\u00ab"},
                                        {"not","\u00ac"},
                                        {"shy","\u00ad"},
                                        {"reg","\u00ae"},
                                        {"macr","\u00af"},
                                        {"deg","\u00b0"},
                                        {"plusmn","\u00b1"},
                                        {"sup2","\u00b2"},
                                        {"sup3","\u00b3"},
                                        {"acute","\u00b4"},
                                        {"micro","\u00b5"},
                                        {"para","\u00b6"},
                                        {"middot","\u00b7"},
                                        {"cedil","\u00b8"},
                                        {"sup1","\u00b9"},
                                        {"ordm","\u00ba"},
                                        {"raquo","\u00bb"},
                                        {"frac14","\u00bc"},
                                        {"frac12","\u00bd"},
                                        {"frac34","\u00be"},
                                        {"iquest","\u00bf"},
                                        {"Agrave","\u00c0"},
                                        {"Aacute","\u00c1"},
                                        {"Acirc","\u00c2"},
                                        {"Atilde","\u00c3"},
                                        {"Auml","\u00c4"},
                                        {"Aring","\u00c5"},
                                        {"AElig","\u00c6"},
                                        {"Ccedil","\u00c7"},
                                        {"Egrave","\u00c8"},
                                        {"Eacute","\u00c9"},
                                        {"Ecirc","\u00ca"},
                                        {"Euml","\u00cb"},
                                        {"Igrave","\u00cc"},
                                        {"Iacute","\u00cd"},
                                        {"Icirc","\u00ce"},
                                        {"Iuml","\u00cf"},
                                        {"ETH","\u00d0"},
                                        {"Ntilde","\u00d1"},
                                        {"Ograve","\u00d2"},
                                        {"Oacute","\u00d3"},
                                        {"Ocirc","\u00d4"},
                                        {"Otilde","\u00d5"},
                                        {"Ouml","\u00d6"},
                                        {"times","\u00d7"},
                                        {"Oslash","\u00d8"},
                                        {"Ugrave","\u00d9"},
                                        {"Uacute","\u00da"},
                                        {"Ucirc","\u00db"},
                                        {"Uuml","\u00dc"},
                                        {"Yacute","\u00dd"},
                                        {"THORN","\u00de"},
                                        {"szlig","\u00df"},
                                        {"agrave","\u00e0"},
                                        {"aacute","\u00e1"},
                                        {"acirc","\u00e2"},
                                        {"atilde","\u00e3"},
                                        {"auml","\u00e4"},
                                        {"aring","\u00e5"},
                                        {"aelig","\u00e6"},
                                        {"ccedil","\u00e7"},
                                        {"egrave","\u00e8"},
                                        {"eacute","\u00e9"},
                                        {"ecirc","\u00ea"},
                                        {"euml","\u00eb"},
                                        {"igrave","\u00ec"},
                                        {"iacute","\u00ed"},
                                        {"icirc","\u00ee"},
                                        {"iuml","\u00ef"},
                                        {"eth","\u00f0"},
                                        {"ntilde","\u00f1"},
                                        {"ograve","\u00f2"},
                                        {"oacute","\u00f3"},
                                        {"ocirc","\u00f4"},
                                        {"otilde","\u00f5"},
                                        {"ouml","\u00f6"},
                                        {"divide","\u00f7"},
                                        {"oslash","\u00f8"},
                                        {"ugrave","\u00f9"},
                                        {"uacute","\u00fa"},
                                        {"ucirc","\u00fb"},
                                        {"uuml","\u00fc"},
                                        {"yacute","\u00fd"},
                                        {"thorn","\u00fe"},
                                        {"yuml","\u00ff"},
                                        // das hier ist HTML 4.0
                                        {"Alpha","\u0391"},
                                        {"Beta","\u0392"},
                                        {"Gamma","\u0393"},
                                        {"Delta","\u0394"},
                                        {"Epsilon","\u0395"},
                                        {"Zeta","\u0396"},
                                        {"Eta","\u0397"},
                                        {"Theta","\u0398"},
                                        {"Iota","\u0399"},
                                        {"Kappa","\u039a"},
                                        {"Lambda","\u039b"},
                                        {"Mu","\u039c"},
                                        {"Nu","\u039d"},
                                        {"Xi","\u039e"},
                                        {"Omicron","\u039f"},
                                        {"Pi","\u03a0"},
                                        {"Rho","\u03a1"},
                                        {"Sigma","\u03a3"}, // der Sprung muss so sein
                                        {"Tau","\u03a4"},
                                        {"Upsilon","\u03a5"},
                                        {"Phi","\u03a6"},
                                        {"Chi","\u03a7"},
                                        {"Psi","\u03a8"},
                                        {"Omega","\u03a9"},
                                        {"alpha","\u03b1"},
                                        {"beta","\u03b2"},
                                        {"gamma","\u03b3"},
                                        {"delta","\u03b4"},
                                        {"epsilon","\u03b5"},
                                        {"zeta","\u03b6"},
                                        {"eta","\u03b7"},
                                        {"theta","\u03b8"},
                                        {"iota","\u03b9"},
                                        {"kappa","\u03ba"},
                                        {"lambda","\u03bb"},
                                        {"mu","\u03bc"},
                                        {"nu","\u03bd"},
                                        {"xi","\u03be"},
                                        {"omicron","\u03bf"},
                                        {"pi","\u03c0"},
                                        {"rho","\u03c1"},
                                        {"sigmaf","\u03c2"},
                                        {"sigma","\u03c3"},
                                        {"tau","\u03c4"},
                                        {"upsilon","\u03c5"},
                                        {"phi","\u03c6"},
                                        {"chi","\u03c7"},
                                        {"psi","\u03c8"},
                                        {"omega","\u03c9"},
                                        {"thetasym","\u03d1"},
                                        {"upsih","\u03d2"},
                                        {"piv","\u03d6"},
                                        // ab hier geht es in Bloecken weiter...
                                        {"bull","\u2022"},
                                        {"hellip","\u2026"},
                                        {"prime","\u2032"},
                                        {"oline","\u203e"},
                                        {"frasl","\u2044"},
                                        {"weierp","\u2118"},
                                        {"image","\u2111"},
                                        {"real","\u211c"},
                                        {"trade","\u2122"},
                                        {"alefsym","\u2135"},
                                        {"larr","\u2190"},
                                        {"uarr","\u2191"},
                                        {"rarr","\u2192"},
                                        {"darr","\u2193"},
                                        {"harr","\u2194"},
                                        {"crarr","\u21b5"},
                                        {"lArr","\u21d0"},
                                        {"uArr","\u21d1"},
                                        {"rArr","\u21d2"},
                                        {"dArr","\u21d3"},
                                        {"hArr","\u21d4"},
                                        {"forall","\u2200"},
                                        {"part","\u2202"},
                                        {"exist","\u2203"},
                                        {"empty","\u2205"},
                                        {"nabla","\u2207"},
                                        {"isin","\u2208"},
                                        {"notin","\u2209"},
                                        {"ni","\u220b"},
                                        {"prod","\u220f"},
                                        {"sum","\u2211"},
                                        {"minus","\u2212"},
                                        {"lowast","\u2217"},
                                        {"radic","\u221a"},
                                        {"prop","\u221d"},
                                        {"infin","\u221e"},
                                        {"ang","\u2220"},
                                        {"and","\u22a5"},
                                        {"or","\u22a6"},
                                        {"cap","\u2229"},
                                        {"cup","\u222a"},
                                        {"int","\u222b"},
                                        {"there4","\u2234"},
                                        {"sim","\u223c"},
                                        {"cong","\u2245"},
                                        {"asymp","\u2248"},
                                        {"ne","\u2260"},
                                        {"equiv","\u2261"},
                                        {"le","\u2264"},
                                        {"ge","\u2265"},
                                        {"sub","\u2282"},
                                        {"sup","\u2283"},
                                        {"nsub","\u2284"},
                                        {"sube","\u2286"},
                                        {"supe","\u2287"},
                                        {"oplus","\u2295"},
                                        {"otimes","\u2297"},
                                        {"perp","\u22a5"},
                                        {"sdot","\u22c5"},
                                        {"lceil","\u2308"},
                                        {"rceil","\u2309"},
                                        {"lfloor","\u230a"},
                                        {"rfloor","\u230b"},
                                        {"lang","\u2329"},
                                        {"rang","\u232a"},
                                        {"loz","\u25ca"},
                                        {"spades","\u2660"},
                                        {"clubs","\u2663"},
                                        {"hearts","\u2665"},
                                        {"diams","\u2666"},
                                        {"ensp","\u2002"},
                                        {"emsp","\u2003"},
                                        {"thinsp","\u2009"},
                                        {"zwnj","\u200c"},
                                        {"zwj","\u200d"},
                                        {"lrm","\u200e"},
                                        {"rlm","\u200f"},
                                        {"ndash","\u2013"},
                                        {"mdash","\u2014"},
                                        {"lsquo","\u2018"},
                                        {"rsquo","\u2019"},
                                        {"sbquo","\u201a"},
                                        {"ldquo","\u201c"},
                                        {"rdquo","\u201d"},
                                        {"bdquo","\u201e"},
                                        {"dagger","\u2020"},
                                        {"Dagger","\u2021"},
                                        {"permil","\u2030"},
                                        {"lsaquo","\u2039"},
                                        {"rsaquo","\u203a"}
                                      };

  protected static Hashtable _specialCharMap=new Hashtable();
  // Array in die Hashtable kopieren
  static
  {
    for (int i=0;i<_specialChars.length;i++)
    {
      String s=_specialChars[i][0];
      char c=_specialChars[i][1].charAt(0);
      _specialCharMap.put(s,new Character(c));
    }
  }
}